from spinecore.common.helpers.contract.metaclasses.singleton import Singleton


class Logger(metaclass=Singleton):
    def __init__(self, run_id: str):
        self.run_id = run_id

    def time_function(self, func, msg: str):
        raise NotImplementedError



from spinecore.common.context.base_context import BaseContext


class IStep:
    async def process(self, ctx: BaseContext, df):
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__class__.__name__



from typing import Union

try:
    from polars import LazyFrame as _LazyFrame, DataFrame as _EagerDataFrame
    PolarsDataFrame = Union[_LazyFrame, _EagerDataFrame]
except ImportError:
    raise ImportError("Polars must be installed")



import networkx as nx

default_dag_style = {
    "node_color": "tab:blue", "alpha": 0.8, "width": 2, "font_size": 12,
    "font_color": "black", "font_weight": "bold", "arrows": True,
    "arrowsize": 25, "bbox": dict(facecolor="skyblue", boxstyle="round", ec="silver", pad=0.3),
}


def topological_pos(graph: nx.DiGraph):
    """Display in topological order, with simple offsetting for legibility"""
    pos_dict = {}
    for i, node_list in enumerate(nx.topological_generations(graph)):
        x_offset = len(node_list) / 2
        y_offset = 0.1
        for j, name in enumerate(node_list):
            pos_dict[name] = (j - x_offset, -i + j * y_offset)

    return pos_dict



from abc import ABC
from typing import List

from spinecore.common.helpers.contract.IStep import IStep


class Workflow(ABC):
    """Abstract class that represents general spine workflow"""

    def __init__(self, name: str, graph):
        self.name = name
        self.graph = graph

    def add_after(self, next_steps: List[IStep], prev_step: List[IStep]):
        """Add step after another"""
        raise NotImplementedError()

    def add_last(self, next_step: IStep):
        """Push step to the end of the workflow"""
        raise NotImplementedError()

    def flat(self):
        """Return flatten representation of the workflow"""
        raise NotImplementedError()

    def get_before(self, step=None):
        """Get the previous step"""
        raise NotImplementedError()

    def get_all_sub_workflows(self) -> List:
        raise NotImplementedError

    def show_workflow(self, style: dict = None) -> None:
        raise NotImplementedError



import asyncio
from typing import List, Dict

from spinecore.common.context.base_context import BaseContext
from spinecore.common.helpers.contract.IExecutor import IExecutor
from spinecore.common.helpers.contract.IStep import IStep
from spinecore.common.helpers.contract.Workflow import Workflow


class AsyncWorkflowExecutor(IExecutor):
    """Async implementation of workflow executor"""

    @staticmethod
    async def __execute_step(ctx: BaseContext, step: IStep, dependencies):
        result = ctx.get_spine_logger().time_function(step.process, f"Step {step.__class__.__name__}")
        return await result(ctx, *dependencies)

    async def __get_dependency_results(self, ctx, dependencies, steps_results: dict, workflow: Workflow) -> List:
        return [await self.__execute_step_by_dependencies(ctx, workflow, steps_results, dependency)
                for dependency in dependencies]

    async def __execute_step_by_dependencies(self, ctx: BaseContext, workflow: Workflow,
                                             steps_results: dict, step: IStep):
        if step in steps_results:
            return steps_results[step]
        dependencies = workflow.get_before(step)
        dependencies_results = await self.__get_dependency_results(ctx, dependencies, steps_results, workflow)
        steps_results[step] = await self.__execute_step(ctx, step, dependencies_results)

    async def __execute_sub_graph(self, ctx, sub_workflow):
        """Execute each graph in sync"""
        steps_results = {}
        flat_graph = sub_workflow.flat()
        for step in flat_graph:
            await self.__execute_step_by_dependencies(ctx, sub_workflow, steps_results, step)
        return steps_results

    async def __execute_all_steps(self, ctx: BaseContext, workflow: Workflow) -> List[Dict]:
        """Execute in async all sub workflows"""
        sub_workflows = workflow.get_all_sub_workflows()
        sub_workflows_results = await asyncio.gather(
            *[asyncio.create_task(self.__execute_sub_graph(ctx, sub_graph)) for sub_graph in sub_workflows])
        return sub_workflows_results

    async def execute(self, ctx: BaseContext, workflow: Workflow) -> List[Dict]:
        ctx.logger.info(f"Executing {workflow.name}")
        return await self.__execute_all_steps(ctx, workflow)
